import numpy as np
from tensorflow import keras
from keras import models
from keras import layers
import copy
from lammps import lammps

def one_hot(n, nmax):
    v = np.zeros(nmax)
    v[n] = 1.
    return v.reshape(1, nmax)

def new_model():
    model = models.Sequential()
    model.add(layers.Dense(HIDDEN_SIZE, input_shape=(SIZE,)))
    model.add(layers.Activation('relu'))
    model.add(layers.Dense(SIZE))
    model.add(layers.Activation('linear'))
    model.compile(loss='mse', optimizer='adam', metrics=['mae'])
    return model

def save_model(episode):
    file_name = 'model_'+ str(episode)+'.h5'
    model.save(file_name)
    print("Model saved to disk")
    return 0

def reset_env(SIZE):
    lmp = lammps()
    lmp.file("intestine2D.in")
    action = np.random.randint(0, SIZE)
    string_action = "variable index equal " + str(action)
    lmp.command(string_action)
    lmp.command("run 10000")
    v_com = lmp.extract_compute("reward",0,1)
    com = v_com[0]
    v_state = one_hot(action, SIZE)
    #after the action, it becomes the state
    return lmp, v_state, com

def step_env(lmp, action, old_com):
    string_action = "variable index equal " + str(action)
    lmp.command(string_action)
    lmp.command("run 10000")
    v_com = lmp.extract_compute("reward",0,1)
    new_com = v_com[0]
    v_state = one_hot(action, SIZE)
    reward = new_com - old_com
    return v_state, new_com, reward

r = 0.
r_sum = 0
file = open('learning.log','w')
SIZE = 10
HIDDEN_SIZE = 20
BATCH_SIZE = 1
MAX_STEPS = 1
EPISODES = 30000
model = new_model()

#execute the q learning
y = 0.5
eps   = 1.0
eps_f = 0.01
decay_factor = (eps_f/eps)**(1./EPISODES)
for i in range(EPISODES):
    lmp, s, com = reset_env(SIZE)
    eps *= decay_factor
    r_sum = 0.
    for j in range(MAX_STEPS):
        if np.random.random() < eps:
            a = np.random.randint(0, SIZE)
        else:
            a = np.argmax(model.predict(s))
        new_s, com, r = step_env(lmp, a, com)
        target = r + y*np.max(model.predict(new_s))
        target_vec = model.predict(s)      
        target_vec[0][a] = target
        model.fit(s, target_vec, epochs=1, batch_size=1, verbose=0)
        s = new_s.copy()
        r_sum += r
    if (i % 1000 == 0 and i > 1):
        save_model(i)
        file.write(str(i) + " " + str(r_sum) + "\n")
        file.flush()

save_model(i)
file.close()